#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2016 Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import collections
import errno
import functools
import importlib
import json
import logging
import multiprocessing.dummy as mp
import os.path
import posixpath
import pyalpm
import re
import shutil
import sqlite3
import subprocess
import urllib.error
import urllib.parse

import MemoizeDB
import XCGF
import XCPF.common
import XCPF.Pkg
import XCPF.PkgInfo
import XCPF.RepoWrapper
import XCPF.SRCINFO



################################## Constants ###################################

PKGBUILD_CACHEDIR = 'pkgbuilds'
PKGBUILD_NAME = 'PKGBUILD'
SRCINFO_NAME = '.SRCINFO'


################################## Functions ###################################

def pkgbuilds_cachepath():
  '''
  Return a standard path for the cached package info.
  '''
  return XCPF.common.cachepath(PKGBUILD_CACHEDIR)



def get_official_pkgbuild_url(repo, pkgbase):
  '''
  Get PKGBUILD URL for the given official pyalpm package.

  Returns None if the package is not an official Arch Linux package.
  '''
  for r, _, _, gitrepo in XCPF.common.ARCHLINUX_OFFICIAL_REPOS:
    if r == repo:
      break
  else:
    logging.warning('{} is not an official repo'.format(repo))
    return None

  pkgbuild_url = '{}{}/{}'.format(
    XCPF.common.ARCHLINUX_ORG_ABS_GIT_HOST,
    XCPF.common.ARCHLINUX_ORG_ABS_GIT_TRUNK.format(
      gitrepo=gitrepo, pkgbase=pkgbase
    ),
    PKGBUILD_NAME
  )
  return pkgbuild_url



def get_pkgbuild_cachepath(repo, pkgbase):
  '''
  Get the path of a locally cached PKGBUILD for the given pkgbase.
  '''
  return os.path.join(pkgbuilds_cachepath(), repo, pkgbase, PKGBUILD_NAME)



def get_official_srcinfo(repo, pkgbase, ttl):
  '''
  Get SRCINFO for the given official pyalpm pkgbase.
  '''
  pkgbuild_url = get_official_pkgbuild_url(repo, pkgbase)
  pkgbuild_path = get_pkgbuild_cachepath(repo, pkgbase)
  dpath = os.path.dirname(pkgbuild_path)
  os.makedirs(dpath, exist_ok=True)
  XCGF.mirror(pkgbuild_url, pkgbuild_path, cache_time=ttl)

  srcinfo_path = os.path.join(dpath, SRCINFO_NAME)
  try:
    srcinfo_mtime = os.path.getmtime(srcinfo_path)
  except FileNotFoundError:
    pass
  else:
    pkgbuild_mtime = os.path.getmtime(pkgbuild_path)
    if srcinfo_mtime > pkgbuild_mtime:
      return XCPF.SRCINFO.Srcinfo(path=srcinfo_path)
  # This should work but it fails if the install script is missing.
#   cmd = ['makepkg', '-p', PKGBUILD_NAME, '--printsrcinfo']
#   srcinfo = subprocess.check_output(cmd, cwd=dpath)
#   with open(srcinfo_path, 'wb') as f:
#     f.write(srcinfo)
#   return XCPF.SRCINFO.Srcinfo(text=srcinfo.decode())
  cmd = ['mksrcinfo']
  subprocess.check_output(cmd, cwd=dpath)
  srcinfo = XCPF.SRCINFO.Srcinfo(path=srcinfo_path)
  for pkgname in srcinfo.pkgnames():
    if pkgname != pkgbase:
      pkgname_link = os.path.join(os.path.dirname(dpath), pkgname)
      XCGF.relink(pkgbase, pkgname_link)
  return srcinfo



def get_official_srcinfo_pkginfo(opi, pkg, ttl, arch):
  '''
  Get package information for the given official pyalpm package from SRCINFO.
  '''
  repo = pkg.db.name
  # The cachedir symlinks repo/pkgname to repo/pkgbase if different.
  symlink_path = os.path.join(pkgbuilds_cachepath(), repo, pkg.name)
  if os.path.exists(symlink_path):
    pkgbase = os.path.basename(os.path.realpath(symlink_path))
  else:
    # This is currently the only way to get the pkgbase. Hopefully this will
    # change in the future.
    pi = opi.pkginfo_by_pyalpm_package(pkg)
    pkgbase = pi['pkgbase']
  srcinfo = get_official_srcinfo(pkg.db.name, pkgbase, ttl)
  return srcinfo.pkginfo(pkg.name, arch=arch)



def get_alpm_pkg_download(cachedir, pkg, urls, more_params=None):
  '''
  Create a download element for a pyalpm package.
  '''
  params = {
    'out': pkg.filename,
    'dir': cachedir,
    'checksum': 'sha-256={}'.format(pkg.sha256sum)
  }
  if more_params:
    params.update(more_params)
  return ('addUri', urls, params)



######################### Additional Package Downloads #########################

def maybe_add_pacserve_downloads(outdir, env, pkgs, downloads=None):
  '''
  Maybe add Aria2 Pacserve downloads for the given packages.

  outdir:
    The output directory for the downloads.

  env:
    The Pacboy environment object.

  downloads:
    A dictionary mapping package names to Aria2 download jobs.

  pkgs:
    A list of tuples. The first item in each tuple is a XCPF.Pkg.Pkg and the
    second is the matching pyalpm package.
  '''
  pacserve_server = env.get(XCPF.RepoWrapper.ENV_KEY_PACSERVE, None)
  if not pacserve_server or not pkgs:
    return downloads
  try:
    import ThreadedServers.Pacserve
  except ImportError as e:
    logging.warning('pacserve server specified in configuration file but ThreadedServers.Pacserve could not be imported: {}'.format(e))
    return downloads

  found_on_pacserve = ThreadedServers.Pacserve.search_pkgs(
    pacserve_server,
    (pkg.filename for _, pkg in pkgs)
  )
  if found_on_pacserve:
    if downloads is None:
      downloads = dict()
    for _, pyalpm_pkg in pkgs:
      try:
        url = found_on_pacserve[pyalpm_pkg.filename]
      except KeyError:
        continue
      else:
        logging.info(
          '{} is available via Pacserve'.format(pyalpm_pkg.filename)
        )
        downloads.setdefault(pyalpm_pkg.name, []).append(
          get_alpm_pkg_download(outdir, pyalpm_pkg, [url])
        )
  return downloads



def maybe_add_ipfs_downloads(outdir, env, pkgs, downloads=None):
  '''
  Maybe add Aria2 IPFS downloads for the given packages.

  Same arguments as maybe_add_pacserve_downloads.
  '''
  if not pkgs:
    return downloads
  localhost = XCGF.recursive_get(
    env,
    (XCPF.RepoWrapper.ENV_KEY_IPFS, 'localhost')
  )
  if not localhost:
    return downloads
  repo = pkgs[0][0].repo()
  arch = pkgs[0][1].arch
  repopath = XCGF.recursive_get(
    env,
    (XCPF.RepoWrapper.ENV_KEY_IPFS, 'repos', repo)
  )
  if not repopath:
    return downloads
  repopath = repopath.replace('$repo', repo).replace('$arch', arch)
  for _, pyalpm_pkg in pkgs:
    urlpath = posixpath.join(repopath, pyalpm_pkg.filename)
    url = urllib.parse.urljoin(localhost, urlpath)
    logging.debug('added {} to download queue'.format(url))
    downloads.setdefault(pyalpm_pkg.name, []).append(
      get_alpm_pkg_download(outdir, pyalpm_pkg, [url])
    )
  return downloads



################################## PyalpmPkg ###################################

class OfficialPyalpmPkg(XCPF.Pkg.PyalpmPkg):

  def __init__(self, pkg, opi):
    '''
    Pkg type for binary pacman packages.

    pkg:
      pyalpm sync package

    opi:
      OfficialPkgInfo object.
    '''
    super(OfficialPyalpmPkg, self).__init__(pkg)
    self.opi = opi
    self.pkginfo = None


  def get_pkginfo(self):
    '''
    Lazily retrieve the package info. This will be the parsed JSON object
    returned by the Arch Linux website.
    '''
    if self.pkginfo is None:
      self.pkginfo = self.opi.pkginfo_by_pyalpm_package(self.pkg)
    return self.pkginfo


  def pkgbase(self):
    pkginfo = self.get_pkginfo()
    return pkginfo['pkgbase']



######################### OfficialBuildExtenderFactory #########################

class OfficialBuildExtenderFactory(XCPF.RepoWrapper.BuildExtenderFactory):
  '''
  BuildExtenderFactory for official packages with ABS support.
  '''
  def __init__(self):
    self.lock = mp.Lock()
    self.supported_repos = list(r[0] for r in XCPF.common.ARCHLINUX_OFFICIAL_REPOS)


  def initialize(self, env):
    self.env = env
    return env


  def extend(self, env, reponame):
    if reponame not in self.supported_repos:
      raise ValueError('{} only supports the official ABS repos: {}'.format(
        self,
        ' '.join(self.supported_repos)
      ))

    for db in env[XCPF.RepoWrapper.ENV_KEY_CPO].handle.get_syncdbs():
      if db.name == reponame:
        try:
          opi = env['opi']
        except KeyError:
          opi = XCPF.PkgInfo.OfficialPkgInfo(env[XCPF.RepoWrapper.ENV_KEY_CPO])
          env['opi'] = opi
          def opi_refresh(*args, **kwargs):
            opi.refresh(*args, **kwargs)
            yield from ()
          env[XCPF.RepoWrapper.ENV_KEY_REFRESH].append(opi_refresh)
        extender = OfficialBuildExtender(env, db.name, self.reflect)
        return env, extender
    else:
      raise ValueError('{} is not enabled'.format(reponame))



  def reflect(self):
    '''
    Get mirrors with reflector. Use a lock to prevent multiple redundant
    queries.
    '''
    # Results could be cached in an attribute and checked here, but then the
    # factory would need a refresh method to clear them.
    reflector_args = self.get_env_config('reflector', 'arguments')

    mirrors = None
    if reflector_args:
      with self.lock:
        try:
          import Reflector
        except ImportError as e:
          logging.warning(
            'Reflector arguments given but the module could not be imported: {}'.format(e)
          )
        else:
          reflector_pargs = Reflector.parse_args(reflector_args)
          try:
            mirrorstatus, mirrors = Reflector.process_options(reflector_pargs)
          except Reflector.MirrorStatusError as e:
            logging.error(str(e))

    # Iterate outside of the lock to prevent consumers from blocking the
    # generator while the lock is still active.
    if mirrors:
      for m in mirrors:
        yield m['url']



############################ OfficialBuildExtender #############################

class OfficialBuildExtender(XCPF.RepoWrapper.BuildExtender):
  '''
  BuildExtender for official packages.
  '''
  def __init__(self, env, repo, reflect):
    '''
    env:
      Same as RepoWrapperFactory.

    repo:
      Name of repo.
    '''
    super(OfficialBuildExtender, self).__init__(env)
    self.repo = repo
    self.reflect = reflect


  def pkg(self, pkg):
    '''
    Return a SrcinfoPkg package if this package is a build target, otherwise
    an OfficialPyalpmPkg.
    '''
    qualified_name = '{}/{}'.format(pkg.db.name, pkg.name)
    if self.env[XCPF.RepoWrapper.ENV_KEY_IS_BUILD_TARGET](qualified_name):
      pkginfo = get_official_srcinfo_pkginfo(
        self.env['opi'],
        pkg,
        self.env[XCPF.RepoWrapper.ENV_KEY_TTL],
        self.env[XCPF.RepoWrapper.ENV_KEY_CPO].arch()
      )
      return XCPF.Pkg.SrcinfoPkg(True, self.repo, pkginfo)
    else:
      return OfficialPyalpmPkg(pkg, self.env['opi'])


  def download_pkgbuilds(self, outdir, pkgbases, with_source_files=False):
    '''
    Download PKGBUILDs and related files using either rsync or parallel
    downloads. Re-use cached PKGBUILDs when possible.
    '''
    pkgbase_jobs = dict()

    if with_source_files:
      # Retrieve the SRCINFO files. This will re-use locally cached objects if
      # possible. The SRCINFO files are created from the PKGBUILDs (which are
      # inherently trusted due to begin part of the official distribution).
      srcinfo_jobs = list(
        XCGF.AsyncJob(
          get_official_srcinfo,
          args=(self.repo, pkgbase, self.env[XCPF.RepoWrapper.ENV_KEY_TTL])
        ) for pkgbase in pkgbases
      )
      # Run the retrieval jobs and load the SRCINFO objects.
      srcinfo_objs = list(XCGF.apply_async_without_id(srcinfo_jobs))
    else:
      srcinfo_objs = None

    # The SRCINFO jobs will cache PKGBUILDs locally. Copy them to the
    # destination directory.
    for pkgbase in pkgbases:
      cached_pkgbuild_path = get_pkgbuild_cachepath(self.repo, pkgbase)
      outpath = os.path.join(outdir, pkgbase, PKGBUILD_NAME)
      try:
        XCGF.copy_if_absent_or_newer(cached_pkgbuild_path, outpath)
      except FileNotFoundError:
        pass

    # Determine if we should use rsync or aria2.
    if self.get_env_config('rsync', 'use', default=False):
      # One rsync job will retrieve all PKGBUILDs and return the qualified
      jobs = [
        XCGF.AsyncJob(
          self.env['opi'].retrieve_abs_via_rsync,
          args=(outdir, {self.repo:pkgbases})
        )
      ]
      if srcinfo_objs:
        for srcinfo in srcinfo_objs:
          jobs.extend(XCPF.RepoWrapper.source_array_download_jobs(
            self.env, srcinfo, os.path.join(outdir, srcinfo.pkgbase())
          ))
      def callback(results):
        yield from results[0]
      grouped_jobs = [{
        'callback': callback,
        'jobs': jobs
      }]
    else:
      grouped_jobs = list()
      # The first item will be the booleans indicating successful downloads of
      # PKGBUILD and related files. The second and beyond will be the source
      # file downloads. The former must succeed. The later may fail and be
      # passed off to makepkg.
      def callback(pkgbase, results):
        if all(results[0]):
          yield '{}/{}'.format(self.repo, pkgbase)

      pkgbase_jobs = dict(
        (pkgbase, [self.get_pkgbuild_download_job(outdir, pkgbase)])
        for pkgbase in pkgbases
      )
      if srcinfo_objs:
        XCPF.RepoWrapper.append_source_download_jobs(
          self.env, outdir, srcinfo_objs, pkgbase_jobs
        )
      grouped_jobs = list(
        {
          'callback': functools.partial(callback, pkgbase),
          'jobs': jobs
        } for pkgbase, jobs in pkgbase_jobs.items()
      )

    return set(), grouped_jobs


  def get_pkgbuild_download_job(self, outdir, pkgbase):
    '''
    Download PKGBUILD for one package base.
    '''
    urls = XCPF.common.search_abs_git_repos(pkgbase, repo=self.repo)
    try:
      url, _ = next(urls)
    except StopIteration:
      return False
    if url is None:
      return False
    logging.info('retrieving {}/{} from {}'.format(self.repo, pkgbase, url))
    downloads = list((
      'addUri',
      [href],
      {
        'out': name,
        'dir': os.path.join(outdir, pkgbase)
      }
    ) for href, name in urls)
    return (self.env[XCPF.RepoWrapper.ENV_KEY_DOWNLOADER], (downloads,))


  def get_additional_mirrors(self, pkgname):
    '''
    Get additional mirrors from Reflector.
    '''
    arch = self.env[XCPF.RepoWrapper.ENV_KEY_CPO].arch()
    for mirror in self.reflect():
      yield mirror.replace('$repo', self.repo).replace('$arch', arch)


############################ AurRepoWrapperFactory #############################

class PacmanRepoWrapperFactory(XCPF.RepoWrapper.RepoWrapperFactory):
  '''
  RepoWrapperFactory for Pacman repos..
  '''
  def __init__(self):
    # The parent class has no explicit __init__ function so we do not need to
    # call it.

    # Map of repo names to BuildExtenderFactory subclasses.
    self.be_factory_classes = dict()

    # Map of BuildExtenderFactory subclasses to instances.
    self.be_factories = dict()


  def initialize(self, env):
    # Keep the build environment.
    self.env = env

    # Clear everything because this may be a re-initialization.
    self.be_factory_classes.clear()
    self.be_factories.clear()

    # Each item in "build extender factories" is a list of strings. For each
    # list (x) the first item is the fully qualifed name of the build extender
    # factory and all following items are repo names supported by that build
    # extender. The order determines preference and the first matching build
    # extender will be used for a given repo. Example x:
    # ['FooBuildExtender', 'core', 'extra', 'community']
    for x in XCGF.recursive_get(
      env,
      (XCPF.RepoWrapper.ENV_KEY_PACBOY_CONF, 'repos', 'build extender factories'),
      default=[]
    ):
      # Expand the loaded repos in x. This will expand "*" to the Pacman repos,
      # for example.
      for repo in env[XCPF.RepoWrapper.ENV_KEY_CPO].expand_repos(x[1:]):
        # Add the build extender class in x to the dictionary if necessary.
        if repo not in self.be_factory_classes:
          self.be_factory_classes[repo] = x[0]
    return env


  def wrap(self, env, reponame):
    # Keep the build environment.
    self.env = env

    # The names of the standard Pacman repos.
    pacman_repos = set(
      db.name for db in env[XCPF.RepoWrapper.ENV_KEY_CPO].handle.get_syncdbs()
    )
    # This repo wrapper only supports Pacman binary repos. Raise an error if
    # something else has been requested.
    if reponame not in pacman_repos:
      raise ValueError(
        '{} only supports enabled Pacman repo'.format(self)
      )

    # Check if there is a build extender factory class for the requested repo.
    try:
      bef_cls = self.be_factory_classes[reponame]
    except KeyError:
      be_obj = None
    else:
      # Check if the build extender factory has already been loaded and
      # initalized
      try:
        # If this succeeds then it has already been loaded and initialized.
        be_factory = self.be_factories[bef_cls]
      except (AttributeError, KeyError):
        # Load it, initialize it and store it in be_factory_classes.
        module, cls_name = bef_cls.rsplit('.', 1)
        # Import the class and call it.
        be_factory = getattr(importlib.import_module(module), cls_name)()
        env = be_factory.initialize(env)
        # The build extender factory can be re-used so store it.
        try:
          self.be_factories[bef_cls] = be_factory
        except AttributeError:
          self.be_factories = {bef_cls: be_factory}

      env, be_obj = be_factory.extend(env, reponame)

    # Create the repo wrapper with the optional build extender. If the build
    # extender is None then it will just act as a regular Pacman repo.
    wrapper = PacmanRepoWrapper(
      self.env, reponame,
      build_extender=be_obj,
    )
    return env, wrapper




############################# OfficialRepoWrapper ##############################

class PacmanRepoWrapper(XCPF.RepoWrapper.RepoWrapper):
  '''
  RepoWrapper for Pacman repos (i.e. ones accessible via pyalpm). This will use
  BuildExtenders to provide build support when available.
  '''

  def __init__(
    self,
    *args,
    build_extender=None,
    additional_pacman_servers=None,
    **kwargs
  ):
    '''
    *args:
      RepoWrapper arguments.

    build_extender:
      Instance of BuildExtender to provide build support.

    additional_pacman_servers:
      Additional servers in the same format as "Server" entries in pacman.conf

    **kwargs:
      RepoWrapper keyword arguments.
    '''
    super(PacmanRepoWrapper, self).__init__(*args, **kwargs)
    self.build_extender = build_extender
    if additional_pacman_servers:
      additional_pacman_servers = list(
        s.replace('$repo', self.name).replace(
          '$arch', self.env[XCPF.RepoWrapper.ENV_KEY_CPO].arch()
        ) for s in additional_pacman_servers
      )
    self.additional_pacman_servers = additional_pacman_servers

    self.db = None
    self.load_db()


  def load_db(self):
    '''
    Load the matching pyalpm database and check if it is official.
    '''
    for db in self.env[XCPF.RepoWrapper.ENV_KEY_CPO].handle.get_syncdbs():
      if db.name == self.name:
        self.db = db
        break
    else:
      raise RepoWrapperError('no such database: {}'.format(self.name), self)


  def alpm_to_xcpf_pkg(self, pkg):
    '''
    Wrap an ALPM package in OfficialPkg.
    '''
    if self.build_extender is not None:
      return self.build_extender.pkg(pkg)
    else:
      return XCPF.Pkg.PyalpmPkg(pkg)


  def refresh(self, force=False, temporary=False):
    '''
    Clear the cached data.
    '''
    self.load_db()
    if self.build_extender is not None:
      yield from self.build_extender.refresh(force=force)


  def satisfy(self, targets, check_groups=True, replacements=False):
    '''
    Search the enabled official databases for the targets.
    '''
    self.load_db()
    groups = dict()
    unsatisfied = set()

    if not targets:
      satisfied = XCPF.Pkg.PkgSet(
        self.alpm_to_xcpf_pkg(p) for p in self.db.pkgcache
      )
      return satisfied, groups, unsatisfied

    satisfied = list()

    ver_reqs = XCPF.common.collect_version_requirements(targets)

    for qualified_name in set(ver_reqs):
      repo, name = XCPF.common.repo_and_package(qualified_name)
      if repo is not None and repo != self.name:
        unsatisfied.add(qualified_name)
        continue

      reqs = list(XCPF.common.get_required_version_strings(ver_reqs, (qualified_name,)))
      pkg = XCPF.common.select_name_or_first(
        name,
        XCPF.common.find_common_satisfiers(self.db.pkgcache, reqs)
      )
      if pkg and (replacements or pkg.name == name):
        satisfied.append(pkg)
        continue
      else:
        # Even if the version requirements are not satisfied, check for a
        # package with the given name. If one exists then it precludes a group
        # of the same name.
        if self.db.get_pkg(name):
          unsatisfied.add(qualified_name)
          continue
        elif check_groups:
          # If there are version requirements then the name can be assumed to be
          # a package and not a group.
          if not ver_reqs[qualified_name]:
            # find_grp_pkgs does not support the qualified name (e.g. core/base)
            grppkgs = XCPF.Pkg.PkgSet(
              self.alpm_to_xcpf_pkg(p) for p in pyalpm.find_grp_pkgs(
                (self.db,), name
              )
            )
            if grppkgs:
              groups[qualified_name] = grppkgs
              continue

      unsatisfied.add(qualified_name)

    satisfied = XCPF.Pkg.PkgSet(
      self.alpm_to_xcpf_pkg(p) for p in satisfied
    )
    return satisfied, groups, unsatisfied


  def upgradable(self, pkgs, downgradable=False):
    self.load_db()
    upgradable = XCPF.Pkg.PkgSet()
    orphans = list()
    dbs = (self.db,)
    for p in pkgs:
      s = pyalpm.sync_newversion(p, dbs)
      if s:
        upkg = self.alpm_to_xcpf_pkg(s)
        upkg.upgrade_of = p
        upgradable.add(upkg)
        continue
      elif downgradable:
        s = self.db.get_pkg(p.name)
        if s and s.version != p.version:
          upkg = self.alpm_to_xcpf_pkg(s)
          upkg.upgrade_of = p
          upgradable.add(upkg)
          continue
      if not self.db.get_pkg(p.name):
        orphans.append(p)
    return upgradable, orphans


  def search(self, terms, by=XCPF.RepoWrapper.SEARCH_BYS[0], intersect=False):
    if by not in XCPF.RepoWrapper.SEARCH_BYS:
      raise NotImplementedError('{}: search by {}'.format(self, by))
    else:
      regexes = list(re.compile(t, flags=re.IGNORECASE) for t in terms)
      check = all if intersect else any
      for p in self.db.pkgcache:
        for f in XCPF.RepoWrapper._search_fields(by, p):
          if check(r.search(f) for r in regexes):
            yield self.alpm_to_xcpf_pkg(p)
            break


  def list(self):
    for p in self.db.pkgcache:
      yield p.name, p.version, p.desc


  def groups(self, names=None, include_members=False):
    if not (names or include_members):
      groups = set()
      for p in self.db.pkgcache:
        groups.update(p.groups)
      return groups
    else:
      groups = dict()
      for p in self.db.pkgcache:
        if p.groups:
          pkg = self.alpm_to_xcpf_pkg(p)
          for g in p.groups:
            if names and g not in names:
              continue
            try:
              groups[g].add(pkg)
            except KeyError:
              groups[g] = XCPF.Pkg.PkgSet(pkgs=(pkg,))
      return groups



  def download_pkgbuilds(self, *args, **kwargs):
    if self.build_extender is None:
      return super(PacmanRepoWrapper, self).download_pkgbuilds(*args, **kwargs)
      raise NotImplementedError('{}: download_pkgbuilds [no BuildExtender]'.format(self))
    else:
      return self.build_extender.download_pkgbuilds(*args, **kwargs)



  def download_packages(self, outdir, pkgs):
    self.load_db()
    # Get the Pyalpm packages.
    pkgs = list(
      x for x in ((p, self.db.get_pkg(p.pkgname())) for p in pkgs)
      if x[1] is not None
    )
    cachedirs = [outdir] + \
      self.env[XCPF.RepoWrapper.ENV_KEY_CPO].pacman_conf.get_option('CacheDir')


    # Check which packages are already in the cache.
    found = XCPF.Pkg.PkgSet()
    for pkg, pyalpm_pkg in pkgs:
      expected_checksum = pyalpm_pkg.sha256sum
      for cachedir in cachedirs:
        path = os.path.join(outdir, pyalpm_pkg.filename)
        checksum = XCGF.get_checksum(path, typ='sha256')
        # No file. Keep scanning.
        if checksum is None:
          continue

        # Expected file is already in the cache.
        elif checksum == expected_checksum:
          logging.debug('found {} in {}'.format(pyalpm_pkg.filename, cachedir))
          found.add(pkg)
          break

        # Incorrect checksum. Delete it and keep scanning.
        else:
          os.remove(path)
          continue

    pkgs = list(x for x in pkgs if x[0] not in found)

    # Copy local files.
    for pkg, pyalpm_pkg in pkgs:
      for url in self.db.servers:
        if url[:7] == 'file://':
          local_path = os.path.join(url[7:], pyalpm_pkg.filename)
          cache_path = os.path.join(outdir, pyalpm_pkg.filename)
          try:
            shutil.copy(local_path, output_path)
          except FileNotFoundError:
            continue
          else:
            checksum = XCGF.get_checksum(output_path, typ='sha256')
            if checksum == pyalpm_pkg.sha256sum:
              logging.debug('copied {} to {}'.format(local_path, cache_path))
              found.add(pkg)
              break
            else:
              logging.debug('bad checksum for {}'.format(local_path))
              os.remove(output_path)

    pkgs = list(x for x in pkgs if x[0] not in found)
    if not pkgs:
      return found, list()

    # If Pacserve is enabled then Pacboy should download available packages from
    # it before trying to download them from the mirrors. What the following
    # does is run a local query to determine which packages are available on
    # Pacserve. Download jobs are created for those packages which will first
    # try to download from Pacserve and then try to download from the mirrors if
    # that fails, while all other packages will de downloaded from the mirrors
    # directly. There will be one asynchronous job per package so everything is
    # still run in parallel, but for each asynchronous job there may be up to 2
    # sequential jobs (Pacserve, then mirrors).
    downloads = dict()
    downloads = maybe_add_pacserve_downloads(
      outdir, self.env, pkgs, downloads=downloads
    )
    downloads = maybe_add_ipfs_downloads(
      outdir, self.env, pkgs, downloads=downloads
    )

    for pkg, pyalpm_pkg in pkgs:
      urls = list(
        os.path.join(url, pyalpm_pkg.filename) for url in self.db.servers
      )
      if self.build_extender:
        urls.extend(self.build_extender.get_additional_mirrors(pyalpm_pkg.name))
      downloads.setdefault(pyalpm_pkg.name, []).append(
        get_alpm_pkg_download(outdir, pyalpm_pkg, urls)
      )

    if not downloads:
      return found, list()

    # Run downloads in parallel in single job. Each job will pass a list of
    # downoads to Aria2 and return a list of booleans indicating which jobs
    # succeeded. Those that succeed are finished. Those that fail may fall back
    # to the second download job for that package (e.g. Pacserve may be tried
    # first, but then the mirrors if Pacserve fails). To do this, the set of
    # successful downloads must be collected and then used to determine which
    # ones failed.
    #
    # If a callback in a job passed to XCGF's async functions yields another
    # XCGF.AsyncJob, it will be re-run in the pool. The callback function below
    # exhausts downloads until one succeeds for each package or there are none
    # left for a package. Each "round" groups all download jobs together to
    # pass them to Aria2 at once.


    downloader = self.env[XCPF.RepoWrapper.ENV_KEY_DOWNLOADER]
    pkgnames = list(downloads.keys())

    def callback(results, pkgnames=pkgnames, downloaded=None, n=0):
      if downloaded is None:
        downloaded = set()
      if results:
        downloaded |= set((p for p,r in zip(pkgnames, results) if r))
      failed = set(pkgnames) - downloaded
      if failed:
        remaining = list(
          (p, downloads[p][n]) for p in failed if downloads[p][n:]
        )
        if remaining:
          # Attempting to use zip above will result in ValueErrors when trying
          # to zip an empty list.
          pkgnames, dls = zip(*remaining)
          return XCGF.AsyncJob(
            downloader,
            args=(dls,),
            result_handler=functools.partial(
              callback,
              pkgnames=pkgnames,
              downloaded=downloaded,
              n=n+1
            )
          )
      return ('{}/{}'.format(self.name, p) for p in downloaded)


    def group_callback(results):
      for r in results:
        yield from r

    grouped_jobs = [{
      'jobs': (callback(None),),
      'callback': group_callback
    }]
    return found, grouped_jobs

