#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2016 Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import collections
import errno
import json
import logging
import MemoizeDB
import os.path
import sqlite3
import subprocess
import urllib.error
import XCGF
import XCPF.common



################################## Functions ###################################

def collect_pkgbases(pkgs):
  '''
  Return a dict mapping pkgbases to lists of Pkgs.
  '''
  pkgbases = dict()
  for pkg in pkgs:
    try:
      pkgbases[pkg.pkgbase].append(pkg)
    except KeyError:
      pkgbases[pkg.pkgbase] = [pkg]
  return pkgbases



############################### PkgError ################################

class PkgError(Exception):
  '''Exceptions raised by methods of the Pkg class.'''
  def __init__(self, msg, error=None):
    self.msg = msg
    self.error = error

  def __str__(self):
    if self.error is None:
      return '{}: {}'.format(self.__class__.__name__, self.msg)
    else:
      return '{}: {} [{}]'.format(self.__class__.__name__, self.msg, self.error)



################################# Pkg #################################

class Pkg(object):
  '''
  A generic interface for package metadata.
  '''

  def __init__(self, build):
    '''
    Subclasses must call this class's __init__ method.

    build:
      True if this package is to be built, False otherwise.
    '''
    self.build = build

    # The build level. This determines if checkdeps are included in all_deps()
    self.with_checkdeps = True

    # True if this package is being installed as a dependency.
    self.is_dep = False

    # True if this package should be ignored. Set for upgrades.
    self.is_ignored = False

    # The local pyalpm package which this package will replace in an upgrade,
    # if any.
    self.upgrade_of = None


  def display_fields(self, extended=False):
    '''
    Return an iterator that can be accepted by XCPF.common.InfoDisplay.display,
    i.e. one that iterates over values of

        <field name>, <field value>, <per line>

    where <per line> indicates if list values should be displayed on separate
    lines. This should begin with Repository, Name, Version, Description and
    Architecture followed by other fields that emulate the output of pacman -Si.

    If "extended" is True, include additional information similar to pacman
    -Sii.
    '''
    return NotImplemented


  def __str__(self):
    '''
    This should not be overridden.
    '''
    return '{}({})'.format(self.__class__.__name__, self.qualified_pkgname())


  ############################# PKGBUILD variables #############################

  def checkdepends(self):
    return NotImplemented

  def conflicts(self):
    return NotImplemented

  def depends(self):
    return NotImplemented

  def groups(self):
    return NotImplemented

  def makedepends(self):
    return NotImplemented

  def pkgbase(self):
    return NotImplemented

  def pkgdesc(self):
    return NotImplemented

  def pkgname(self):
    return NotImplemented

  def provides(self):
    return NotImplemented

  def replaces(self):
    return NotImplemented


  ################################### Other ####################################

  def maintainers(self):
    '''
    Return a list of names of the maintainers of this package, i.e. the names
    that the user must trust to trust this package. For example, for Pacman repo
    packages this should be the maintainers and/or packagers and for the AUR
    this should be the maintainers.
    '''
    return list()

  def rid(self):
    '''
    Return a unique release identifier for this package. This may be anything
    that is tied to the particular release of the package such as a checksum or
    a modification time. It should change whenever the package or its metadata
    changes. It will be used to track trust in particular releases.

    This must return a string.
    '''
    return self.version()

  def version(self):
    '''
    Return the package version (<pkgver>-<pkgrel>).
    '''
    return NotImplemented

  def repo(self):
    '''
    Return the package repository or a suitable origin identifier.
    '''
    return NotImplemented

  def qualified_pkgbase(self):
    '''
    Returned the pkgbase with the repo prefix.
    '''
    return '{}/{}'.format(self.repo(), self.pkgbase())

  def qualified_pkgname(self):
    '''
    Returned the pkgname with the repo prefix.
    '''
    return '{}/{}'.format(self.repo(), self.pkgname())

  def operation_deps(self, alldeps=False):
    '''
    Iterate over the dependencies required for the given operation. All runtime
    deps are yielded. Other deps will be yielded as a function of the Pkg's
    "build" attribute, or the "alldeps" keyword argument.

    alldeps:
      If True, yield all dependendencies (runtime, build, check).
    '''
    yield from self.depends()
    if self.build or alldeps:
      yield from self.makedepends()
      if self.with_checkdeps or alldeps:
        yield from self.checkdepends()



################################## SrcinfoPkg ##################################

class SrcinfoPkg(Pkg):
  '''
  Pkg subclass wrapper around XCPF.SRCINFO.Pkginfo
  '''

  def __init__(self, build, repo, pkginfo):
    '''
    repo: The name of the repository.

    pkginfo: An instance of XCPF.SRCINFO.Pkginfo
    '''
    super(SrcinfoPkg, self).__init__(build)
    self.reponame = repo
    self.pkginfo = pkginfo


  def display_fields(self, extended=False):
    yield 'Repository', self.reponame, False
    if extended:
      yield 'Package Base', self.pkginfo.pkgbase, False
    yield 'Name', self.pkginfo.pkgname, False
    yield 'Version', self.pkginfo.version, False
    yield 'Description', self.pkginfo.pkgdesc, False
    yield 'Architecture', self.pkginfo.arch, False
    yield 'URL', self.pkginfo.url, False
    yield 'Licenses', self.pkginfo.licenses, False
    yield 'Groups', self.pkginfo.groups, False
    yield 'Provides', self.pkginfo.provides, False
    yield 'Depends On', self.pkginfo.depends, False
    if extended:
      yield 'Makedepends On', self.pkginfo.makedepends, False
      yield 'Checkdepends On', self.pkginfo.checkdepends, False
    yield 'Optional Deps', self.pkginfo.optdepends, True
    yield 'Conflicts With', self.pkginfo.conflicts, False
    yield 'Replaces', self.pkginfo.replaces, False


  def checkdepends(self):
    return self.pkginfo.checkdepends

  def conflicts(self):
    return self.pkginfo.conflicts

  def depends(self):
    return self.pkginfo.depends

  def groups(self):
    return self.pkginfo.groups

  def makedepends(self):
    return self.pkginfo.makedepends

  def pkgbase(self):
    return self.pkginfo.pkgbase

  def pkgdesc(self):
    return self.pkginfo.pkgdesc

  def pkgname(self):
    return self.pkginfo.pkgname

  def provides(self):
    return self.pkginfo.provides

  def replaces(self):
    return self.pkginfo.replaces

  def version(self):
    return self.pkginfo.version

  def repo(self):
    return self.reponame



################################## PyalpmPkg ###################################

class PyalpmPkg(Pkg):
  def __init__(self, pkg):
    '''
    Pkg type for binary pacman packages.

    pkg:
      pyalpm sync package
    '''
    super(PyalpmPkg, self).__init__(False)
    self.pkg = pkg

  def display_fields(self, extended=False):
    return XCPF.common.display_fields(self.pkg, extended=extended)

  def conflicts(self):
    return self.pkg.conflicts

  def depends(self):
    return self.pkg.depends

  def groups(self):
    return self.pkg.groups

  def pkgdesc(self):
    return self.pkg.desc

  def pkgname(self):
    return self.pkg.name

  def provides(self):
    return self.pkg.provides

  def replaces(self):
    return self.pkg.replaces

  def pkgname(self):
    return self.pkg.name

  def version(self):
    return self.pkg.version

  def description(self):
    return self.pkg.desc

  def repo(self):
    return self.pkg.db.name

  def maintainers(self):
    return [self.pkg.packager]



#################################### PkgSet ####################################

class PkgSet(object):
  '''
  Sets of Pkgs with normal set operations.
  '''
  def __init__(self, pkgs=None):
    self.pkgs = dict()
    if pkgs:
      for pkg in pkgs:
        self.pkgs[pkg.qualified_pkgname()] = pkg

  def __repr__(self):
    return 'PkgSet({})'.format(repr(self.pkgs))

  def copy(self):
    return self.__class__(self)

  def clear(self):
    self.pkgs.clear()

  def add(self, pkg):
    self.pkgs[pkg.qualified_pkgname()] = pkg
    return self

  def update(self, pkgs):
    for pkg in pkgs:
      self.add(pkg)
    return self

  def __iadd__(self, other):
    return self.update(other)

  def __add__(self, other):
    copy = self.copy()
    return copy.__iadd__(other)

  def remove(self, pkg):
    if isinstance(pkg, str):
      name = pkg
    else:
      name = pkg.qualified_pkgname()
    try:
      del self.pkgs[name]
    except KeyError:
      pass
    return self

  def remove_many(self, pkgs):
    for pkg in pkgs:
      self.remove(pkg)
    return self

  def __isub__(self, other):
    return self.remove_many(other)

  def __sub__(self, other):
    copy = self.copy()
    return copy.__isub__(other)

  def remove_version(self, name, version):
    try:
      if self.pkgs[name].version() == version:
        del self.pkgs[name]
    except KeyError:
      pass
    return self

  def names(self):
    return self.pkgs.keys()

  def ignore(self, names, groups):
    names = set(names)
    groups = set(groups)
    ignored = set()
    for name, pkg in self.pkgs.items():
      if names and name in names:
        ignored.add(name)
      else:
        try:
          if (groups & set(pkg.groups())):
            ignored.add(name)
        except KeyError:
          pass
    for i in ignored:
      del self.pkgs[i]
    return self

  def __and__(self, other):
    copy = self.copy()
    return copy.__iand__(other)

  def __iand__(self, other):
    delenda = set(self.pkgs) - set(other.pkgs)
    for name in delenda:
      del self.pkgs[name]
    return self

  def __or__(self, other):
    copy = self.copy()
    return copy.__ior__(other)

  def __ior__(self, other):
    self.pkgs.update(other.pkgs)
    return self

  def __contains__(self, pkg):
    if isinstance(pkg, str):
      return pkg in self.pkgs
    else:
      return pkg.qualified_pkgname() in self.pkgs

  def __iter__(self):
    yield from self.pkgs.values()

  def __len__(self):
    return len(self.pkgs)

  def __bool__(self):
    return bool(self.pkgs)

  def __getitem__(self, k):
    return self.pkgs[k]

  def __setitem__(self, k, v):
    self.pkgs[k] = v

  def __delitem__(self, k):
    del self.pkgs[k]



################################ Build Ordering ################################

def find_satisfiers(pkgs, req):
  '''
  Find packages that satisfy the given requirement.

  pkgs:
    PkgSet of packages to check.

  req:
    Requirement string as found in a dependency array, e.g. foo>=1.4
  '''
  repo, name = XCPF.common.repo_and_package(req)
  name, rel, ver = XCPF.common.split_version_requirement(name)
  for pkg in pkgs:
    if repo and pkg.repo() != repo:
      continue
    if pkg.pkgname() == name:
      if XCPF.common.satisfies_version_requirement(pkg.version(), rel, ver):
        yield pkg
        continue
    for provision in pkg.provides():
      # Provides should never include a repo prefix so no need to strip it.
      p_name, p_rel, p_ver = XCPF.common.split_version_requirement(provision)
      if name == p_name:
        # If there is no version information assume that it satisfies the req.
        if rel is None or p_rel is None:
          yield pkg
          break
        # If the provided version is fixed, check if it satisfies the req.
        elif p_rel == '=':
          if XCPF.common.satisfies_version_requirement(p_ver, rel, ver):
            yield pkg
            break
        # If the required version is fixed, check if the provision satisfies it.
        elif rel == '=':
          if XCPF.common.satisfies_version_requirement(ver, p_rel, p_ver):
            yield pkg
            break
        # If both involve inequalities in the same direction then the req should
        # be satisfied.
        elif (('>' in p_rel and '>' in rel) or ('<' in p_rel and '<' in rel)):
          yield pkg
        # The inequalities oppose each other.
        else:
          # sgn(p_ver - ver)
          vc = pyalpm.vercmp(p_ver, ver)
          # foo>=1.4 and foo<=1.4
          if vc == 0:
            if '=' in p_rel and '=' in rel:
              yield pkg
              break
          # foo<=1.5 and foo>=1.4
          elif vc > 0:
            if '<' in p_rel:
              yield pkg
              break
          # foo>=1.4 and foo<=1.5
          else:
            if '>' in p_rel:
              yield pkg
              break



class DependencyGraphNode(object):
  '''
  Callable dependency graph node. Each call will propagate dependency levels to
  all nodes above this one in the graph.

  graph: A dictionary object.

  pkg: A Pkg object.
  '''

  def __init__(self, graph, pkgcache, pkg):
    self.graph = graph
    self.pkgcache = pkgcache

    self.pkg = pkg
    self.n = 0
    graph[pkg.qualified_pkgname()] = self
    self.deps = tuple(pkg.operation_deps())
    self.calling = False

  def __call__(self, n):
    # Prevent circular recursion.
    if not self.calling:
      self.calling = True
      self.n = max(self.n, n)
      for d in self.deps:
        for satisfier in find_satisfiers(self.pkgcache, d):
          try:
            self.graph[satisfier.qualified_pkgname()](n+1)
          except KeyError:
            pass
      self.calling = False

  def __str__(self):
    return '{}:{:d}'.format(self.pkg.qualified_pkgname(), self.n)

  def __repr__(self):
    return 'DependencyGraphNode({})'.format(self.__str__())



class DependencyGraph(object):
  '''
  Dependency graph calculator.
  '''
  def __init__(self, pkgs=None):
    self.pkgcache = PkgSet()
    if pkgs:
      self.initialize(pkgs)


  def initialize(self, pkgs):
    '''
    Initialize the nodes.
    '''
    n = len(pkgs)
    self.graph = dict()
    for pkg in pkgs:
      self.pkgcache.add(pkg)
      DependencyGraphNode(self.graph, self.pkgcache, pkg)
    for v in self.graph.values():
      # This has the effect of installing sync packages of equivalent dependency
      # depth first.
      if v.pkg.build:
        v(0)
      else:
        # This could just as well be 1.
        v(n)


  def __iter__(self):
    '''
    Iterate over the DependencyGraphNode objects in installation order.
    '''
    yield from sorted(self.graph.values(), key=lambda x: (x.n, not x.pkg.build, x.pkg.pkgname()), reverse=True)
