#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2016 Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import collections
import errno
import json
import logging
import MemoizeDB
import os.path
import sqlite3
import subprocess
import urllib.error
import XCGF
import XCPF.common
import XCPF.Pkg



################################## Constants ###################################

PKGINFO_DBNAME = 'pkginfo.sqlite3'



################################## Functions ###################################

def pkginfo_version(pkginfo):
  '''
  Format the version string from pkginfo.
  '''
  try:
    return '{}-{}'.format(pkginfo['pkgver'], pkginfo['pkgrel'])
  except KeyError as e:
    raise XCPF.Pkg.PkgError(
      'KeyError for {}'.format(pkginfo['pkgname']),
      error=e
    )



############################### OfficialPkgInfo ################################

def pkginfo_dbpath():
  '''
  Return a standard path for the cached package info.
  '''
  return XCPF.common.cachepath(PKGINFO_DBNAME)



class OfficialPkgInfo(object):
  '''
  A class to handle package information retrieval. It provides optional
  transparent caching of the remote data.

  cpo:
    CommonPacmanOptions

  ttl:
    Time-to-live of cached data.

  testing:
    Retrieve testing data. This applies when querying pkgnames without repo data.

  dbpath:
    Path of caching database.

  mdb:
    MemoizeDB instance.

  dbs:
    pyalpm sync databases.

  resolve_pkgbases:
    Retrieve remote data to resolve pkgbases when querying pkgnames.
  '''
  def __init__(
    self,
    cpo,
    ttl=XCPF.common.DEFAULT_TTL,
    testing=False,
    dbpath=None,
    mdb=None,
    dbs=None,
    resolve_pkgbases=False
  ):
    self.cpo = cpo
    self.resolve_pkgbases = resolve_pkgbases

    self.ttl = ttl
    self.testing = testing

    if dbs is None:
      dbs = cpo.handle.get_syncdbs()
    self.dbs = dbs

    if ttl > 0 or mdb is not None:
      if mdb is None:
        if dbpath is None:
          dbpath = pkginfo_dbpath()
        def f(urls):
          for url in urls:
            logging.info('retrieving {}'.format(url))
            yield url, (XCGF.text_from_url(url),)
        glue = {
          'request' : (f, (('data', 'TEXT'),), ttl)
        }
        conn = sqlite3.connect(
          dbpath,
          detect_types=(sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES)
        )
        mdb = MemoizeDB.MemoizeDB(conn, glue)
        mdb.db_initialize()

      def memoized_archlinux_org_pkg_info(repo, arch, pkgname):
        url = XCPF.common.archlinux_org_pkg_info_url(repo, arch, pkgname)
        txt = mdb.get_one('request', url)
        return json.loads(txt)

      self.retrieve_pkginfo = memoized_archlinux_org_pkg_info

    else:
      self.retrieve_pkginfo = archlinux_org_pkg_info

    self.mdb = mdb


  def arch(self):
    '''
    Get the target architecture.
    '''
    return self.cpo.arch()



  def refresh(self, force=False):
    '''
    Clear cached data.
    '''
    if self.mdb is not None:
      self.mdb.db_clean(wipe=True)



  def pkginfo_by_pkgname(self, pkgname, repo=None):
    '''
    Retrieve package information by pkgname. This will search through the
    official repos until the package is found or the repos are exhausted.
    '''
    arch = self.arch()
    for trepo, a, testing, gr in ARCHLINUX_OFFICIAL_REPOS:
      if (repo is None or repo == trepo) \
      and (self.testing or not testing) \
      and (a == 'any' or a == arch):
        try:
          return self.retrieve_pkginfo(trepo, arch, pkgname)
        except urllib.error.HTTPError as e:
          if e.code == 404:
            continue
          else:
            raise e
    return None



  def pkginfo_by_pyalpm_package(self, pkg):
    '''
    Retrieve package information for an instance of pyalpm.Package.
    '''
    try:
      return self.retrieve_pkginfo(pkg.db.name, pkg.arch, pkg.name)
    except urllib.error.HTTPError as e:
      if e.code == 404:
        return None
      else:
        raise e



  def pkginfo_by_pkgname_or_pyalpm_package(self, pkg):
    '''
    Retrieve package information by pkgname or pyalpm.Package.
    '''
    if isinstance(pkg, pyalpm.Package):
      return self.pkginfo_by_pyalpm_package(pkg)
    else:
      p = self.pyalpm_package(pkg)
      if p is None:
        return self.pkginfo_by_pkgname(pkg)
      else:
        return self.pkginfo_by_pyalpm_package(p)



  def pyalpm_package(self, pkgname):
    '''
    Search databases for a package name.
    '''
    if isinstance(pkgname, pyalpm.Package):
      return pkgname
    else:
      return find_pkgname(self.cpo.handle.get_syncdbs(), pkgname)



  def pyalpm_package_and_pkginfo(self, pkgname):
    '''
    Attempt to retrieve the pyalpm.Package and the package information.
    '''
    pkg = self.pyalpm_package(pkgname)
    if pkg is None:
      pkginfo = self.pkginfo_by_pkgname(pkgname)
    else:
      pkginfo = self.pkginfo_by_pyalpm_package(pkg)
    return pkg, pkginfo



  def map_to_pkgbases(self, pkgs):
    '''
    Iterate over tuples mapping pkgnames to pkgbases. If the pkginfo could not
    be found, then the pkgbase will be None.
    '''
    for p in pkgs:
      pkginfo = self.pkginfo_by_pkgname_or_pyalpm_package(p)
      try:
        yield p, pkginfo['base']
      except (TypeError, KeyError):
        yield p, None



  def map_to_pkgname_pkgbase_repo(self, pkgname):
    '''
    Attempt to map a pkgname or pyalpm.Package to a pkgname, repo and pkgbase.

    Use this to collect package bases for ABS retrieval.
    '''
    pkg = self.pyalpm_package(pkgname)
    if pkg is None:
      name = pkgname
      pkginfo = self.pkginfo_by_pkgname(pkgname)
      try:
        base = pkginfo['pkgbase']
      except (TypeError, KeyError):
        base = None
      try:
        repo = pkginfo['repo']
      except (TypeError, KeyError):
        repo = None
    else:
      name = pkg.name
      repo = pkg.db.name
      pkginfo = self.pkginfo_by_pyalpm_package(pkg)
      try:
        base = pkginfo['pkgbase']
      except (TypeError, KeyError):
        base = None
    return name, repo, base



  def collect_pkgbases_by_repo(self, pkgnames):
    '''
    Attempt to map packages to package bases and organize them by repo. If
    resolve_pkgbases is True then this will retrieve the necessary information
    to correctly map packages to their package bases.

    Returns a dict mapping repos to pkgbases and a dict mapping pkgbases to
    pkgnames.
    '''
    repos = dict()
    pkgbases = dict()
    if self.resolve_pkgbases:
      for name, repo, base in (
        self.map_to_pkgname_pkgbase_repo(p) for p in pkgnames
      ):
        try:
          pkgbases[base].add(name)
        except KeyError:
          pkgbases[base] = set((name,))

        if base is not None and name != base:
          logging.debug('mapped {} to {} in {}'.format(
            name, base, repo
          ))
        if base is None:
          base = name
        try:
          repos[repo].add(base)
        except KeyError:
          repos[repo] = set((base,))
    else:
      repos[None] = set(
        p.name if isinstance(p, pyalpm.Package) else p for p in pkgnames
      )
    return repos, pkgbases



  def retrieve_abs_via_git(self, output_dir, pkgnames):
    '''
    Retrieve the requested files from ABS via the online Git interface if
    possible. Iterates over the found packages and their URLs.

    Use XCPF.common.retrieve_abs_via_git directly if the pkgbases are already
    known.
    '''
    if self.resolve_pkgbases:
      repos, pkgbase_map = self.collect_pkgbases_by_repo(pkgnames)
    else:
      repos = {
        None : set(pkgnames)
      }
      pkgbase_map = dict()
    for repo, pkgbases in repos.items():
      for pkgbase in pkgbases:
        url = XCPF.common.retrieve_abs_via_git(output_dir, pkgbase, repo=repo)
        if url:
          try:
            for pkgname in pkgbase_map[pkgbase]:
              yield pkgname, url
          except KeyError:
            yield pkgbase, url



  def get_blind_abs_rsync_paths(self, repos=None):
    '''
    Get a list of ABS rsync paths for the given architecture and repos.
    '''
    for repo, a, is_testing, foo in ARCHLINUX_OFFICIAL_REPOS:
      if repos and repo not in repos:
        continue
      elif is_testing and not self.testing:
        continue
      elif a == 'any':
        yield ARCHLINUX_ORG_ABS_RSYNC_PATH.format(arch=a, repo=repo)
        if self.arch() != a:
          yield ARCHLINUX_ORG_ABS_RSYNC_PATH.format(arch=self.arch(), repo=repo)
      elif a == self.arch():
        yield ARCHLINUX_ORG_ABS_RSYNC_PATH.format(arch=a, repo=repo)



#   def get_abs_rsync_paths(self, repos):
#     '''
#     Get a list of ABS rsync paths for the given architecture and repos.
#     '''
#     for repo, pkgbases in repos.items():
#       for pkgbase in pkgbases:
#         r = ARCHLINUX_ORG_ABS_RSYNC_PATH.format(arch=self.arch, repo=repo)
#         yield '{}{}/'.format(r, pkgbase)



  def retrieve_abs_via_rsync_by_pkgnames(
    self,
    output_dir,
    pkgnames,
  ):
    '''
    Retrieve the requested files from ABS via rsync.
    '''
    repos, pkgbase_map = self.collect_pkgbases_by_repo(pkgnames)
    for pkgbase in self.retrieve_abs_via_rsync(output_dir, repos):
      try:
        yield from pkgbase_map[pkgbase]
      except KeyError:
        yield pkgbase


  def retrieve_abs_via_rsync(
    self,
    output_dir,
    repos_and_pkgbases
  ):
    '''
    Retrieve ABS files for the given pkgbases and save them in the output
    directory in subfolders for each pkgbase.

    output_dir:
      The output directory.

    repos_and_pkgbases:
      A mapping of repos to pkgbases.
    '''
    for repo, pkgbases in repos_and_pkgbases.items():
      options = list('--include={}/***'.format(x) for x in pkgbases)
      options.append('--exclude=/*')

      if repo is None:
        paths = list(self.get_blind_abs_rsync_paths())
        if paths:
          paths[0] = XCPF.common.ARCHLINUX_ORG_ABS_RSYNC_HOST + abs_paths[0]

      else:
        root = XCPF.common.ARCHLINUX_ORG_ABS_RSYNC_PATH.format(
          arch = self.arch(),
          repo = repo
        )
        paths = [XCPF.common.ARCHLINUX_ORG_ABS_RSYNC_HOST + root]

      yield from self.run_rsync_cmd(output_dir, pkgbases, options, paths)



  def run_rsync_cmd(
    self,
    output_dir,
    pkgbases,
    options,
    paths,
    rsync_path='rsync',
  ):
    '''
    Run the rsync command.

    output_dir:
      The output directory.

    pkgbases:
      The expected pkgbases.

    options:
      The rsync options (e.g. --include..., --excluse...)

    paths:
      The rsync paths.

    rsync_path:
      The path to the rsync executable.
    '''
    cmd = [rsync_path, '-rt']
    cmd.extend(options)
    cmd.extend(paths)
    cmd.append(output_dir)

    # Use mtimes to determine if rsync retrieved a file.
    # This avoids piping and parsing rsync output.
    pkgdirs = tuple(os.path.join(output_dir, x) for x in pkgbases)
    mtimes = dict()
    for d in pkgdirs:
      try:
        mtime = os.path.getmtime(d)
        mtimes[d] = mtime
        os.utime(d, (0,0))
      except OSError as e:
        if e.errno != errno.ENOENT:
          raise e

    logging.debug('invoking {}'.format(cmd))
    with open(os.devnull, 'wb') as f:
      subprocess.call(cmd, stdout=f)

    for d in pkgdirs:
      try:
        mtime = os.path.getmtime(d)
      except OSError as e:
        if e.errno != errno.ENOENT:
          raise e
      else:
        if mtime != 0:
          head, pkgname = os.path.split(d)
          head, repo = os.path.split(head)
          yield '{}/{}'.format(repo, pkgname)
        else:
          t = mtimes[d]
          os.utime(d, (t,t))
