#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2016 Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import os.path
import pyalpm
import re
import subprocess

import XCGF
import XCPF.common



################################## Constants ###################################

# Possible search methods.
SEARCH_BYS = (
  'name-desc', # Package name and description.
  'name', # Package name.
)



######################## Key Constants for Environment #########################

ENV_KEY_CONFIG = 'config' # 3rd-party configuration.
ENV_KEY_CPO = 'cpo' # CommonPacmanOptions object.
ENV_KEY_DATABASE_DIR = None # If set, repowrapper metadata should be temporarily stored in a sub-directory of this directory.
ENV_KEY_DOWNLOADER = 'downloader' # Aria2JsonRpc.Aria2JsonRpc.polymethod_download_bool or equivalent.
ENV_KEY_FOREIGN_CONFIG = 'foreign config' # Nested configuration objects. See RepoWrapperBase.get_env_config()
ENV_KEY_IS_BUILD_TARGET = 'is build target' # Function to determine build targets.
ENV_KEY_PACBOY_CONF = 'pacboy conf' # Pacboy configuration object.
ENV_KEY_PACSERVE = 'pacserve' # Optional Pacserve data.
ENV_KEY_IPFS = 'ipfs' # Optional IPFS data.
ENV_KEY_REFRESH = 'refresh' # List of refresh functions.
ENV_KEY_TTL = 'ttl' # Time-to-live of cached metadata.


################################## Functions ###################################

# This will need to be updated when SEARCH_BYS changes.
# pkg is a pyalpm package.
def _search_fields(by, pkg):
  '''
  For a given search-by method, yield the values of the pyalpm package which
  should be searched
  '''
  yield pkg.name
  yield from pkg.groups
  if by == 'name-desc' and pkg.desc:
    yield pkg.desc



def interpolate_vcs_download_command(cmd, outdir, uris):
  '''
  Interpolate the "VCS download command" in ENV_KEY_PACBOY_CONF with the given
  output directory and URIs.

  cmd:
    The command in the Pacboy configuration for downloading PKGBUILD VCS
    sources. This should contain "%OUTDIR% where the output directory should
    be interpolated, and "%URI% where the list of URIs should be inserted.

  outdir:
    The output directory.

  uris:
    The URIs as they appear in the PKGBUILD, with any name prefix (<name>::) or
    VCS scheme prefix (<vcs>+<uri scheme>, e.g. "git+http").
  '''
  for word in cmd:
    if word == '%URI%':
      yield from uris
    else:
      yield word.replace('%OUTDIR%', outdir)



def source_array_download_jobs(env, srcinfo, outdir):
  '''
  Return jobs that can be passed to XCGF.async_apply to download remote files
  from the source array in the given Srcinfo. This will use the environment's
  ENV_KEY_DOWNLOADER and the VCS downloader specified in ENV_KEY_PACBOY_CONF.

  env:
    The environment passed around between RepoWrappers.

  srcinfo:
    The Srcinfo instance from which to read the source array.

  outdir:
    The output directory in which to save the files.
  '''
  downloader = env[ENV_KEY_DOWNLOADER]
  vcs_download_cmd = env[ENV_KEY_PACBOY_CONF]['build']['VCS download command']

  vcs_uris = list()
  downloads = list()

  for uri, stripped_uri, scheme, name, checksums in srcinfo.remote_sources(
    arch=env[ENV_KEY_CPO].arch()
  ):
    if scheme in ('bzr', 'git', 'hg', 'svn'):
      vcs_uris.append(uri)

    elif scheme in ('http', 'https', 'ftp'):
      params = {
        'out': name,
        'dir': outdir
      }
      # Only one checksum can be passed so pass the strongest one.
      for i in (512, 384, 256, 224, 1, None):
        key = 'md5sums' if i is None else 'sha{:d}sums'.format(i)
        try:
          checksum = checksums[key]
        except KeyError:
          continue
        typ = ('md5' if i is None else 'sha-{:d}'.format(i))
        params['checksum'] = '{}={}'.format(typ, checksum)
        break
      downloads.append(('addUri', [stripped_uri], params))

    else:
      logging.warning('source_array_download_jobs: skipping unsupported scheme: {}'.format(
        scheme
      ))

  if vcs_uris:
    vcs_download_cmd = list(interpolate_vcs_download_command(
      vcs_download_cmd,
      outdir,
      vcs_uris
    ))
    # TODO
    # If --quiet, silence output here.
    yield XCGF.AsyncJob(subprocess.run, args=(vcs_download_cmd,))

  if downloads:
    # The list of downloads is the first argument to downloader. This must be
    # wrapped in another list representing the positional arguments of which
    # this is only one: the downloads list.
    yield XCGF.AsyncJob(downloader, args=(downloads,))



def append_source_download_jobs(env, outdir, srcinfo_objs, pkgbase_jobs):
  '''
  Append source download jobs for the given Srcinfo objects.

  env:
    The environment passed around between RepoWrappers.

  outdir:
    The base output directory. The sources will be downloaded to the
    subdirectory <pkgbase>.

  srcinfo_objs:
    The iterable of Srcinfo objects.

  pkgbase_jobs:
    A dictionary mapping pkgbase strings to lists of jobs.

  Returns pkgbase_jobs updated with the appended source download jobs.
  '''
  for srcinfo in srcinfo_objs:
    jobs = source_array_download_jobs(
      env,
      srcinfo,
      os.path.join(outdir, srcinfo.pkgbase())
    )
    pkgbase_jobs.setdefault(srcinfo.pkgbase(), []).extend(jobs)
  return pkgbase_jobs



def attempt_downloads(downloader, required, dlqueue, callback=None):
  '''
  Attempt lists of downloads until one completes. This can be used to ensure
  that database files are downloaded from the same server and fall back to
  another server if any required download fails, rather than keep successful
  files from one server and complete the missing downloads from another.

  download:
    Function that accepts the same download arguments as
    Aria2RpcManager.download

  required:
    A list of booleans which should be commensurate with each list in the
    download queue.

  dlqueue:
    An itertools.deque of downloads to try where each item should be a list
    of downloads with the same length as "required". The format of each item
    is the same as that accepted by the passed download function.

  callback:
    If given, it will be called with True if the download has succeeded or False
    if not. The return value will be returned.
  '''
  len_req = len(required)
  complete = False
  while dlqueue:
    downloads = dlqueue.popleft()
    complete = True
    dls = downloader(downloads)
    if len(downloads) > len_req:
      zipper = itertools.zip_longest(dls, required, fillvalue=False)
    else:
      zipper = zip(dls, required)
    # Let the loop complete to exhaust the iterator.
    for success, req in zipper:
      if req:
        if not success:
          complete = False
    if complete:
      break
  if callback is None:
    return complete
  else:
    return callback(complete)



############################### RepoWrapperError ###############################

class RepoWrapperError(Exception):
  def __init__(self, msg, wrapper, error=None):
    '''
    Common error raised by RepoWrapper.

    msg: The error message.
    wrapper: The RepoWrapper class that generated the error.
    error: An Exception associated with the error, e.g. an HTTP error.
    '''
    self.msg = msg
    self.wrapper = wrapper
    self.error = error

  def __str__(self):
    if self.error is None:
      return 'RepoWrapperError({}): {}'.format(self.wrapper, self.msg)
    else:
      return 'RepoWrapperError({}): {} [{}]'.format(
        self.wrapper, self.msg, self.error
      )



################################## Base Class ##################################

class RepoWrapperBase(object):
  '''
  Base class with common methods for

  * RepoWrapper
  * RepoWrapperFactory
  * BuildExtender
  * BuildExtenderFactory

  '''

  def __str__(self):
    '''
    Should not be overridden.
    '''
    return XCGF.qualname(self)


  def get_env_config(self, *keys, env=None, default=None):
    '''
    Wrapper around XCGF.recursive_get to retrieve data from this object's
    nested configuration in the environment object.
    '''
    name = str(self)
    if name[-7:] == 'Factory':
      name = name[:-7]
    return XCGF.recursive_get(
      self.env if env is None else env,
      (ENV_KEY_FOREIGN_CONFIG, name, *keys),
      default=default
    )



############################## RepoWrapperFactory ##############################

class RepoWrapperFactory(RepoWrapperBase):
  '''
  Base class to handle configuration and shared object initialization when
  generating RepoWrappers.
  '''
  def initialize(self, env):
    '''
    Initialize resources as necessary.

    env:
      A dictionary of common information and objects that need to be shared
      between RepoWrappers. If a necessary resource does not yet exist then it
      should be created and added to the environment. Data that is only common
      to a subclass of RepoWrapperFactory should be kept in a nested dictionary
      under the key of the class name to avoid polluting the namespace of the
      global environment. For example, if FooRepoWrapperFactory requires the
      variables "a", "b" and "c" then they should be added as

          env['FooRepoWrapperFactory'] = {'a':..., 'b':..., 'c':...}

    Return the possibly updated environment.
    '''
    return env


  def wrap(self, env, reponame):
    '''
    Return the RepoWrapper for the given repo name.

    env:
      Same as "env" in initialize().

    reponame:
      The name of the repo to wrap.

    Return the environement and a subclass instance of RepoWrapper.

    Raise a ValueError if the requested reponame is not supported.
    '''
    return env, NotImplemented


  def finalize(self, env):
    '''
    Handle any cleanup that may be necessary.

    env:
      Same as "env" in initialize().
    '''
    return env



################################## RepoWrapper ##################################

class RepoWrapper(RepoWrapperBase):
  '''
  Template class to provide a generic interface for building Arch Linux packages
  from various sources. Packages returned from the following methods should be
  instances of an associated subclass of Pkg.

  Each instance should be associated with a single Pacman repository so that the
  user's prefered order may be respected.
  '''
  def __init__(self, env, name):
    '''
    Subclasses must invoke this method.

    env:
      Same as RepoWrapperFactory.
    '''
    self.env = env
    self.name = name


  def refresh(self, force=False):
    '''
    Refresh package metadata. This should NOT download the pacman sync database
    associated with the wrapped repo but rather the metadata associated with
    this wrapper. This refresh method will be run after the pacman sync database
    has been updated. It should reload the updated databases and flush any
    cached data to ensure that subsequent operations get "fresh" metadata.

    If the environment variable ENV_KEY_DATABASE_DIR is set, then all metadata
    should be saved to a subdirectory of that directory, UNLESS the metadata has
    no effect on subsequent operations.

    force:
      If True then force a refresh even if cached data is up-to-date. This is
      invoked with "-Syy".

    For IO-bound operations such as downloads, this should yield dictionaries
    that can be passed to XCGF.apply_async_with_grouped_jobs. The callback and
    error_callback of each job, if given, will be called from the main thread
    and so can be used to update the wrapper with the results of the call.
    Alternatively, the common callback for each job can also be used to check
    that all downloads have succeeded and do the same.

    Raise a RepoWrapperError if the operation fails.
    '''
    yield from ()


  def satisfy(self, targets, check_groups=True, replacements=False):
    '''
    Get Pkg objects that satisfy the given targets. A target is satisfied
    if the name and version requirements are match OR if a package replaces the
    target. Functions are provided in XCPF.common (e.g.
    collect_version_requirements) to test dependency satisfaction.

    Targets will be strings of the form

        [<repo>/]<name>[<version specification]

    where bracked ([]) denote optional components. Any target that specifies
    a repo other than the one implemented by the RepoWrapper should be returned
    as unsatisfied. It will be passed on to another RepoWrapper.

    This must support qualified group names as targets, e.g. "core/base".

    If "check_groups" is True, search for groups with names matching <name> in
    addition to matching package names.

    If "replacements" is True, a target may be satisfied by a package that
    replaces it.

    Returns

      * a PkgSet of satisfied targets
      * a dictionary mapping satisfied groups to PkgSets of members
      * the unsatisfied targets as a set of strings

    All hits should be returned without checking if the hits are ignored by the
    current configuration. Ignored groups and packages will be handled
    elsewhere. All members of a group in the repo should be returned. Members
    will be selected elsewhere.

    If targets is empty, the returned PkgSet should include all packages
    available in the repos unless returning them would involve significant
    overhead, in which case the PkgSet should be empty. The other returned
    values should be empty in both cases.
    '''
    return XCPF.Pkg.PkgSet(), dict(), set(targets)


  def upgradable(self, pkgs, downgradable=False):
    '''
    Return a PkgSet of upgradable packages and a list of orphans for the given
    pyalpm packages. Orphans are packages that do not exist in the database.

    pkgs:
      An iterable over local pyalpm packages.

    downgradable:
      If True, for each local package that is not upgradable, check if there is
      a package in the repo with the same name but an earlier version. If so,
      include it in the set of upgradable packages.

    For each upgradable package, set the "upgrade_of" attribute to the local
    pyalpm package that it will replace.

    The default implementation uses satisfy() with appropriate arguments. It
    should be overridden only if it can be implemented more efficiently.
    '''
    installed = dict((i.name, i) for i in pkgs)
    upgradable, groups, orphans = self.satisfy(
      installed, check_groups=False, replacements=True
    )
    uptodate = set()

    for p in sorted(upgradable, key=lambda x: x.pkgname()):
      # This will fail for replacements.
      try:
        local = installed[p.pkgname()]
      except KeyError:
        # This should only occur for replacements.
        for r in p.replaces:
          try:
            local = installed[r]
          except KeyError:
            continue
          else:
            break
      else:
        comp = pyalpm.vercmp(local.version, p.version())
        if not (comp < 0 or (downgradable and comp > 0)):
          uptodate.add(p.qualified_pkgname())
          continue
      p.upgrade_of = local
    upgradable.remove_many(uptodate)
    return upgradable, list(installed[o] for o in orphans)



  def search(self, terms, by=SEARCH_BYS[0], intersect=False):
    '''
    Search for packages. This should be a generator (i.e. return hits in a loop
    using "yield"). It must be iterable.

    terms:
      The search terms.

    by:
      The search method. See the comments for SEARCH_BYS for details.

    intersect:
      If True, all search terms must be matched by each package. If False, only
      one term must match for each package.

    Yield each matching package as a Pkg. The results should be sorted.

    When matching terms against package names, the same terms should also be
    matched against groups.
    '''
    raise NotImplementedError('{}: search'.format(self))


  def list(self):
    '''
    Iterate over packages in the repo:

        <name>, <version>, <description>

    If iterating over all of the packages would involve significant overhead,
    this may return an empty iterator but not a partial list, i.e. it must
    either iterate over all packages or none.

    If <version> or <description> data cannot be retrieved without significan
    overhead then either may be None.

    The packages should be sorted by name.
    '''
    raise NotImplementedError('{}: list'.format(self))


  def groups(self, names=None, include_members=False):
    '''
    List groups and members.

    names:
      The target groups.

    include_members:
      A boolean indicating if

    If names and include_members both evaluate to False then return a set of
    group names, else return a dictionary mapping group names to PkgSets of
    group members. If names is None, return all groups, otherwise only the
    groups listed in names.
    '''
    if not (names or include_members):
      return set()
    else:
      return dict()


  def download_packages(self, outdir, pkgs):
    '''
    Download binary packages to the target directory, if supported.

    outdir:
      The output directory.

    pkgs:
      The Pkgs to download.

    Returns (PkgSet, list). where PkgSet contains the requested Pkgs which have
    already been downloaded and the list contains dictionaries that can be
    passed to XCGF.apply_async_with_grouped_jobs.

    The callback should return the qualified pkgnames of the downloaded package
    for the associated jobs. In the simplest case where a single job is required
    per package (e.g. one download per package), the dictionary should be

        {
          "jobs": [<XCPF.AsyncJob>],
          "callback": lambda x: return [<qualified pkgname>]
        }

    In the case where a single job returns multiple package (e.g. via rsync),
    a list containing each retrieved package should be returned.
    '''
    return XCPF.Pkg.PkgSet(), list()


  def download_pkgbuilds(self, outdir, pkgbases, with_source_files=False):
    '''
    Download PKGBUILDs and related files for the given pkgbases to a determined
    subpath of the output directory.

    outdir:
      The output directory. The files for each pkgbase should be downloaded to
      <outdir>/<pkgbase>.

    pkgbases:
      An iterable of pkgbase strings. These should be unqualified, i.e. there
      should be no repo prefix.

    with_source_files:
      If True, download source files if possible. If this is not possible,
      issue a warning with the logging module.

    Returns (set, list), where the set contains the qualified pkgbases that have
    already been downloaded and the list contains dictionaries that can be
    passed to XCGF.apply_async_with_grouped_jobs.

    The callback should return the qualified pkgbase of the downloaded packages.
    In the simplest case where a single job is required, the dictionary should
    be

        {
          "jobs": [<XCPF.AsyncJob>],
          "callback": lambda x: return [<qualified pkgbase>]
        }

    In the case where one job retrieves the PKGBUILD and another job retrieves
    the source files, the dicitonary should be

        {
          "jobs": [
            <XCPF.AsyncJob pkgbuild job>,
            <XCPF.AsyncJob source files job>
          ],
          "callback": lambda x: return [<qualified pkgbase>]
        }
    '''
    return set(), list()




############################# BuildExtenderFactory #############################

class BuildExtenderFactory(RepoWrapperBase):
  '''
  Base class to handle configuration and shared object initialization when
  generating BuildExtenders.
  '''
  def initialize(self, env):
    '''
    Initialize resources as necessary.

    env:
      Same as RepoWrapperFactory

    Return the possibly updated environment.
    '''
    return env


  def extend(self, env, reponame):
    '''
    Return the BuildExtender for the given repo name.

    env:
      Same as "env" in initialize().

    reponame:
      The name of the repo to extend.

    Return the environement and a subclass instance of BuildExtender.
    '''
    return env, NotImplemented


  def finalize(self, env):
    '''
    Handle any cleanup that may be necessary.

    env:
      Same as "env" in initialize().
    '''
    return env



################################ BuildExtender #################################

class BuildExtender(RepoWrapperBase):
  '''
  Add build support to PacmanRepoWrapper.
  '''
  def __init__(self, env):
    self.env = env


  def refresh(self, force=False):
    '''
    Refresh metadata. See notes for RepoWrapper.refresh().
    '''
    yield from ()


  def pkg(self, pkg):
    '''
    Wrap an alpm package in a Pkg subclass that provides all necessary
    metadata for dependency resolution.
    '''
    raise NotImplementedError('{}: pkg'.format(self))


  def get_additional_mirrors(self, pkgname):
    '''
    Get additional mirrors at runtime if possible. This is used to append
    Reflector mirrors to the official package downloads.

    Iterate over the mirrors. Interpolate "$repo" and "$arch" as necessary.
    '''
    yield from ()


  def download_pkgbuilds(self, outdir, pkgbases, with_source_files=False):
    '''
    Same as RepoWrapper.download_pkgbuilds.
    '''
    raise NotImplementedError('{}: get_pkgbuilds'.format(self))
