#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2015  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import os.path
import urllib.parse

import XCGF



################################### Globals ####################################

PKGBASE_STRING = 'pkgbase'
PKGNAME_STRING = 'pkgname'

ARRAYS = (
  'arch',
  'backup',
  'checkdepends',
  'conflicts',
  'depends',
  'groups',
  'license',
  'makedepends',
  'md5sums',
  'noextract',
  'optdepends'
  'options'
  'options',
  'pkgname',
  'provides',
  'replaces',
  'sha1sums',
  'sha224sums',
  'sha256sums',
  'sha384sums',
  'sha512sums',
  'source',
  'validpgpkeys',
)

MULTI_ARCH = (
  'source',
  'depends', 'checkdepends', 'makedepends', 'optdepends',
  'provides', 'conflicts', 'replaces',
  'md5sums', 'sha1sums', 'sha224', 'sha256sums', 'sha384sums', 'sha512sums',
)



################################## Functions ###################################

def insert_pkgbase(pkginfo, pkgbase):
  '''
  Insert items from the pkgbase.
  '''
  for bk, bv in pkgbase.items():
    if bk not in pkginfo:
      pkginfo[bk] = bv.copy()
    else:
      pkginfo[bk].extend(bv)
  return pkginfo



################################### PkgInfo ####################################

class PkgInfo(object):
  def __init__(self, srcinfo, arch=None):
    self.srcinfo = srcinfo
    self.arch = arch


  def __getattr__(self, name):
    if name in ('pkgname', 'pkgbase'):
      return self.srcinfo[name]
    elif name in ARRAYS:
      return list(self.generate(name))
    else:
      if name == 'version':
        try:
          return '{}-{}'.format(
            self.srcinfo['pkgver'][0], self.srcinfo['pkgrel'][0]
          )
        except KeyError:
          return None
      try:
        return self.srcinfo[name][0]
      except KeyError:
        return None


  def generate(self, name):
    '''
    Iterate over array values including architecture-specific values.
    '''
    try:
      yield from self.srcinfo[name]
    except KeyError:
      pass
    if self.arch is not None and name in MULTI_ARCH:
      try:
        yield from self.srcinfo['{}_{}'.format(name, self.arch)]
      except KeyError:
        pass


  def __iter__(self):
    yield from self.srcinfo


  def __contains__(self, x):
    return (x in self.srcinfo)



################################### Srcinfo ####################################

class Srcinfo(object):
  '''
  .SRCINFO file parser. This will extract all data
  '''
  def __init__(self, text=None, path=None, url=None):
    self.info = dict()
    if text is not None:
      self.load_from_text(text)
    elif path is not None:
      self.load_from_path(path)
    elif url is not None:
      self.load_from_url(url)



  def parse(self, lines):
    '''
    Parse the lines of a .SRCINFO file.

    This will return a dictionary with the following structure:

        {
          PKGBASE_STRING : [
            <pkgbase string>,
            <pkgbase dictionary mapping pkgbuild variables to lists>
          ],
          PKGNAME_STRING : {
            <pkgname_1> : <dictionary of pkgname_1 values>,
            <pkgname_2> : <dictionary of pkgname_2 values>,
            ...
          }
        }
    '''
    self.info.clear()
    obj = None

    for line in lines:
      stripped_line = line.strip()

      if not stripped_line or stripped_line[0] == '#':
        continue

      k, v = stripped_line.split(' = ', 1)

      if line[0] == '\t':
        try:
          obj[k].append(v)
        except KeyError:
          obj[k] = [v]

      else:
        obj = dict()
        if k == PKGBASE_STRING:
          self.info[PKGBASE_STRING] = (v, obj)
        else:
          try:
            self.info[k][v] = obj
          except KeyError:
            self.info[k] = {v : obj}


  def load_from_text(self, text):
    '''
    Parse .SRCINFO text.
    '''
    self.parse(text.split('\n'))


  def load_from_path(self, path):
    '''
    Load a local .SRCINFO file.
    '''
    logging.debug('loading SRCINFO from {}'.format(path))
    with open(path, 'r') as f:
      self.parse(f)


  def load_from_url(self, url):
    '''
    Load a remote .SRCINFO file.
    '''
    logging.debug('loading SRCINFO from {}'.format(url))
    text = XCGF.text_from_url(url)
    self.load_from_text(text)


  def load_from_pkgbuild(self, pkgbuild_path, srcinfo_path=None):
    '''
    Extract SRCINFO from a local PKGBUILD. This should only be used with trusted
    PKGBUILDs.

    pkgbuild_path:
      The path to the PKGBUILD.

    srcinfo_path:
      An optional path to a .SRCINFO file. If given , it is interpretted
      relative to pkgbuild_path. If this path exists and is not older than the
      PKGBUILD then SRCINFO will be extracted from it. The generated SRCINFO
      will also be saved to this path if it is set.
    '''
    pkgbuild_path = os.path.abspath(pkgbuild_path)
    dpath, name = os.path.split(path)
    if srcinfo_path is not None:
      srcinfo_path = os.path.join(dpath, srcinfo_path)
      try:
        srcinfo_mtime = os.path.getmtime(srcinfo_path)
      except FileNotFoundError:
        pass
      else:
        pkgbuild_mtime = os.path.getmtime(pkgbuild_path)
        if pkgbuild_mtime < srcinfo_mtime:
          logging.debug('reloading SRCINFO from {}'.format(pkgbuild_path))
          with open(srcinfo_path, 'r') as f:
            return self.load_from_text(f.read())

    logging.debug('extracting SRCINFO from {}'.format(pkgbuild_path))
    cmd = ('makepkg', '-p', name, '--printsrcinfo')
    text = subprocess.check_output(cmd, cwd=dpath)

    if srcinfo_path is not None:
      with open(srcinfo_path, 'w') as f:
        f.write(text)

    self.load_from_text(text)


  def pkgbase(self):
    '''
    Return the pkgbase name.
    '''
    try:
      return self.info[PKGBASE_STRING][0]
    except KeyError:
      return None


  def pkgnames(self):
    '''
    Iterate over pkgnames in SRCINFO.
    '''
    try:
      yield from self.info[PKGNAME_STRING]
    except KeyError:
      pass


  def pkgbaseinfo(self, arch=None):
    '''
    Return a PkgInfo object containing the pkgbase info.
    '''
    try:
      info = self.info[PKGBASE_STRING][1]
    except KeyError:
      return None
    else:
      return PkgInfo(info, arch=arch)


  def pkginfo(self, pkgname, arch=None):
    '''
    Return package information for one of the packages within the SRCINFO.
    This will copy the data in pkgbase and update it with values in the section
    for this package. The returned object will be a dictionary mapping PKGBUILD
    variables to lists of values.

    Two non-list values will also be included: PKGBASE_STRING and PKGNAME_STRING
    '''
    try:
      pkginfo = self.info[PKGNAME_STRING][pkgname].copy()
      try:
        pkgbase = self.info[PKGBASE_STRING][0]
        pkginfo = insert_pkgbase(pkginfo, self.info[PKGBASE_STRING][1])
      except KeyError:
        pkgbase = pkgname
        pass
    except KeyError:
      return None
    pkginfo[PKGBASE_STRING] = pkgbase
    pkginfo[PKGNAME_STRING] = pkgname
    return PkgInfo(pkginfo, arch=arch)



  def remote_sources(self, arch=None):
    '''
    Parse the source array and iterate over remote source downloads. For each
    remote source, yield the following tuple:

      (<uri>, <stripped uri>, <scheme>, <output name>, <checksums>)

    where

      <uri>:
        The URI as it appears in the source array.

      <parsed uri>.
        The URI with any "<name>::" and "<vcs>+" prefixes stripped.

      <scheme>:
        Either the VCS prefix of the scheme if there was one (e.g. "git" in
        "git+http") or the scheme itself.

      <output name>:
        The name under which the resource should be saved. If this is included
        as a prefix with the "<name>::" notation then it will be used, otherwise
        this will be the basename of the URI's path component.

      <checksums>:
        Any checksums associated with this source. This will be a dictionary
        mapping PKGBUILD array names to string values.

    arch:
      The target architecture.
    '''
    pkgbaseinfo = self.pkgbaseinfo(arch=arch)
    for i, src in enumerate(pkgbaseinfo.source):
      # Check for the name specifier.
      try:
        name, url = src.split('::', 1)
      except ValueError:
        url = src
        name = None

      split_url = urllib.parse.urlsplit(url)

      # Skip local sources.
      if not split_url.scheme:
        continue

      # Determine the effective scheme, which may be a VCS prefix
      # (e.g. git+https).
      try:
        scheme, real_scheme = split_url.scheme.split('+', 1)
      except ValueError:
        scheme = split_url.scheme
      else:
        # SplitResult objects are just glorified tuples and thus immutable.
        # split_url.scheme = real_scheme
        split_url = urllib.parse.SplitResult(real_scheme, *split_url[1:])

      if name is None:
        name = split_url.path.rsplit('/', 1)[-1]

      # Determine checksums;
      checksums = dict()
      for j in (512, 384, 256, 224, 1, None):
        key = 'md5sums' if j is None else 'sha{:d}sums'.format(j)
        try:
          value = getattr(pkgbaseinfo, key)[i]
        except (KeyError, IndexError):
          continue
        else:
          if value == 'SKIP':
            continue
        checksums[key] = value

      yield src, urllib.parse.urlunsplit(split_url), scheme, name, checksums

