#!/usr/bin/bash
set -e

# This is an example snippet set to test parallel package building with Pacboy.
# While ordering the builds, Pacboy tracks each package's level in the
# dependency graph. Instead of building each package sequentially, it will start
# all builds on the same level in the background. When the level inscreases, it
# will wait for the previous level packages to finish and install them before
# proceeding. This should ensure that all dependencies are satisfied for the
# next level. To use this snippet set, add the following entries to the pacboy
# configuration file:
#
# "build": {
#   "snippets": {
#     "prologue": "snippets/example-parallel_build-prologue.sh",
#     "epilogue": "snippets/example-parallel_build-epilogue.sh",
#     "build": "snippets/example-parallel_build-build.sh",
#     "pacman": "snippets/example-parallel_build-pacman.sh"
#   }
# }



# Array to store PIDs of backgrounded build processes.
MAKEPKG_PIDS=()

# Temporary script in which to accumulate pending installation commands.
PENDING_MAKEPKG_CMDS_SCRIPT="$(mktemp)"


# Clear the array of PIDs and pending commands script.
function reinit_pending()
{
  MAKEPKG_PIDS=()
  cat > "$PENDING_MAKEPKG_CMDS_SCRIPT" <<'SCRIPT'
#!/usr/bin/bash
set -e

SCRIPT
}

# Wait for all pending builds to complete, then install the packages and reset
# the queue.
function wait_and_install_pending()
{
  for PID_ in "${MAKEPKG_PIDS[@]}"
  do
    # Wait for all pending builds to finish.
    while true
    do
      echo "waiting for PID $PID_ to finish:"
      # Use "&&" so that it only sleeps if the file still exists.
      if cat /proc/$PID_/cmdline 2>/dev/null
      then
        sleep 5
      else
        break
      fi
    done
  done
  # Run the accumulated installation commands.
  bash "$PENDING_MAKEPKG_CMDS_SCRIPT"
  # Clear the commands for the next level.
  reinit_pending
}

function finish()
{
  wait_and_install_pending
  rm "$PENDING_MAKEPKG_CMDS_SCRIPT"
}

# Use this if you want to always ensure that "finish" is run.
# trap finish EXIT
